/* Dump Utility of Magnetic Tape on UNIX system               */
/*                                 Copyright(c)1990 by T.Toda */
/* Original Version with submodules controlling BIOS directly */
/*           for PC9801 has been coded by T.Toda in 1986-1988 */
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>

#define	BUFSIZE	65536

#define	min(a,b)	((a)<(b)?(a):(b))

static	unsigned char	buffer[BUFSIZE];
static	int	fhd;
static	struct	mtop	mtopbuf;
static	struct	mtget	mtgetbuf;
static	int	ebcdic = 1;

static	unsigned char	ea_table[] = {
	0x00, 0x01, 0x02, 0x03, 0x9c, 0x09, 0x86, 0x7f,
	 0x97, 0x8d, 0x8e, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
	0x10, 0x11, 0x12, 0x13, 0x9d, 0x85, 0x08, 0x87,
	 0x18, 0x19, 0x92, 0x8f, 0x1c, 0x1d, 0x1e, 0x1f,
	0x80, 0x81, 0x82, 0x83, 0x84, 0x0a, 0x17, 0x1b,
	 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x05, 0x06, 0x07,
	0x90, 0x91, 0x16, 0x93, 0x94, 0x95, 0x96, 0x04,
	 0x98, 0x99, 0x9a, 0x9b, 0x14, 0x15, 0x9e, 0x1a,
	0x20, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
	 0xa8, 0xa9, 0x5b, 0x2e, 0x3c, 0x28, 0x2b, 0x21,
	0x26, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xa0,
	 0xb0, 0xb1, 0x5d, 0x24, 0x2a, 0x29, 0x3b, 0x5e,
	0x2d, 0x2f, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,
	 0xb8, 0xb9, 0x7c, 0x2c, 0x25, 0x5f, 0x3e, 0x3f,
	0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1,
	 0xc2, 0x60, 0x3a, 0x23, 0x40, 0x27, 0x3d, 0x22,

	0xc3, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	 0x68, 0x69, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
	0xca, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70,
	 0x71, 0x72, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0,
	0xd1, 0x7e, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
	 0x79, 0x7a, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
	 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
	0x7b, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	 0x48, 0x49, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed,
	0x7d, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50,
	 0x51, 0x52, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3,
	0x5c, 0x9f, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
	 0x59, 0x5a, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	 0x38, 0x39, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff};

static	char	*hex2();
static	void	dumps(), dump1();
char	*getenv();

void	main(argc,argv,envp)
int	argc;
char	*argv[], *envp[];
{
	int	eargc = argc;
	char	**eargv = argv;
	char	*devname, *ptr;

	/* Procession of the switches */
	while(( eargc > 1 ) && ( eargv[1][0] == '-' )){
		for( ptr = eargv[1]+1 ; *ptr ; ptr++ ){
			switch( *ptr ){
			case 'a':
				ebcdic = 0;
				break;
			case 'e':
				ebcdic = 1;
				break;
			default:
				fputs("Usage: ",stdout);
				fputs(argv[0],stdout);
				fputs(" [-ae] [device file name]\n",stdout);
				exit(8);
			}
		}
		eargc--;	eargv++;
	}

	/* Open the Device File */
	if( eargc > 2 ){
		fputs(argv[0],stdout);
		fputs(" : too many arguments\n",stdout);
		exit(8);
	}
	else if( eargc == 2 ){
		devname = eargv[1];
	}
	else{
		devname = getenv("TAPE");
		if( devname == NULL )devname = DEFTAPE;
	}

	fhd = open(devname,O_RDONLY);
	if( fhd == -1 ){
		fputs(devname,stdout);
		fputs(" : Device Open Error\n",stdout);
		exit(16);
	}

	/* Main Loop */
	for(;;){
		fputs("Physical File Skip Value ? ",stdout);
		gets(buffer);
		mtopbuf.mt_count = atol(buffer);
		if( mtopbuf.mt_count > 0 ){
			mtopbuf.mt_op = MTFSF;
			ioctl(fhd,MTIOCTOP,&mtopbuf);
		}
		else if( mtopbuf.mt_count < 0 ){
			mtopbuf.mt_count = - mtopbuf.mt_count;
			mtopbuf.mt_op = MTBSF;
			ioctl(fhd,MTIOCTOP,&mtopbuf);
		}

		fputs("Block(Physical Record) Skip Value ? ",stdout);
		gets(buffer);
		mtopbuf.mt_count = atol(buffer);
		if( mtopbuf.mt_count > 0 ){
			mtopbuf.mt_op = MTFSR;
			ioctl(fhd,MTIOCTOP,&mtopbuf);
		}
		else if( mtopbuf.mt_count < 0 ){
			mtopbuf.mt_count = - mtopbuf.mt_count;
			mtopbuf.mt_op = MTBSR;
			ioctl(fhd,MTIOCTOP,&mtopbuf);
		}

		dumps();
		do{
			fputs("Continue(Y/N)? ",stdout);
			gets(buffer);
			if((buffer[0] == 0)||
			   (buffer[0] == 'N')||(buffer[0] == 'n'))exit(0);
		}while((buffer[0] != 'Y')&&(buffer[0] != 'y'));
	}
}

/* Reading and Displaying Processes */
static	void	dumps()
{
	int	len;
	unsigned	lc;

	for(;;){
		lc = 0;
		len = read(fhd,buffer,BUFSIZE);
		ioctl(fhd,MTIOCGET,&mtgetbuf);
		printf("Block No. %ld in File No. %ld\n",
			mtgetbuf.mt_blkno,mtgetbuf.mt_fileno);
		if( len ){
			for( lc=0 ; lc<len ; lc+=16 ){
				dump1(buffer+lc,min(len-lc,16));
			}
			printf("Block Length = %d\n",len);
		}
		else{
			fputs("## Tape Mark Block ##\n",stdout);
		}
		do{
			fputs("Next Block(Y/N)? ",stdout);
			gets(buffer);
			if((buffer[0] == 0)||
			   (buffer[0] == 'N')||(buffer[0] == 'n'))return;
		}while((buffer[0] != 'Y')&&(buffer[0] != 'y'));
	}
}

/* Display one Line */
static	void	dump1(buffer,len)
unsigned char	*buffer;
int	len;
{
	int	i, cntl, cntl0;
	unsigned char	c;

	for( i=0 ; i<16 ; i++ ){
		fputs( i<len ? hex2(buffer[i]) : "  " ,stdout);
		putchar(' ');
	}
	fputs("  ",stdout);
	cntl0 = 2;
	for( i=0 ; i<len ; i++ ){
		c = buffer[i];
		if( ebcdic )c = ea_table[c];
		if(c < ' '){
			c += '@';
			cntl = 1;
		}
		else if(c == 0x7f){
			c = '?';
			cntl = 1;
		}
		else if(c > 0x7f){
			c = ' ';
			cntl = 1;
		}
		else{
			cntl = 0;
		}
		if(cntl != cntl0){
			fputs( cntl ? "\033[7m" : "\033[0m" ,stdout);
			cntl0 = cntl;
		}
		putchar(c);
	}
	fputs("\033[0m\n",stdout);
}

/* Make	Hexadecimal Code from 1-Byte Character */
static	char	*hex2(c)
int	c;
{
	static	char	hexnum[] = "0123456789ABCDEF";
	static	char	result[3];

	c &= 0x00ff;
	result[0] = hexnum[c >> 4];
	result[1] = hexnum[c & 0x000f];
	result[2] = 0;
	return result;
}
