/* MT Locating Utility for Labelled Tape on UNIX system       */
/*                                 Copyright(c)1990 by T.Toda */
/* Original Version with submodules controlling BIOS directly */
/*          for PC-9801 has been coded by T.Toda in 1986-1988 */
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>

#define	EOMSTOP	0
/* Equate to 1 if the Device never let the Tape   */
/*  to go through the second Tape Mark at the EOM */ 
/* Equate to 0 if the Device let the Tape         */
/*  to read the second Tape Mark and require      */
/*  two `backspaces' to read the last data block  */

#define	TMSKIP	1
/* Equate to 1 if the Tape Handler treats        */
/*  'File Skip' as if 'Tape Mark Skip' so that   */
/*  the 'Backspacing' operation will locate to   */
/*  the end of the destinate file                */
/* Equate to 0 if the Tape Handler distinguishes */
/*  'File Skip' from 'Tape Mark Skip' so that    */
/*  the 'Backspacing' operation will locate to   */
/*  the head of the destinate file               */

static	int	fhd;
static	struct	mtop	mtopbuf;
static	unsigned char	label1[81], label2[81];

static	unsigned char	ea_table[] = {
	0x00, 0x01, 0x02, 0x03, 0x9c, 0x09, 0x86, 0x7f,
	 0x97, 0x8d, 0x8e, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
	0x10, 0x11, 0x12, 0x13, 0x9d, 0x85, 0x08, 0x87,
	 0x18, 0x19, 0x92, 0x8f, 0x1c, 0x1d, 0x1e, 0x1f,
	0x80, 0x81, 0x82, 0x83, 0x84, 0x0a, 0x17, 0x1b,
	 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x05, 0x06, 0x07,
	0x90, 0x91, 0x16, 0x93, 0x94, 0x95, 0x96, 0x04,
	 0x98, 0x99, 0x9a, 0x9b, 0x14, 0x15, 0x9e, 0x1a,
	0x20, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
	 0xa8, 0xa9, 0x5b, 0x2e, 0x3c, 0x28, 0x2b, 0x21,
	0x26, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xa0,
	 0xb0, 0xb1, 0x5d, 0x24, 0x2a, 0x29, 0x3b, 0x5e,
	0x2d, 0x2f, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,
	 0xb8, 0xb9, 0x7c, 0x2c, 0x25, 0x5f, 0x3e, 0x3f,
	0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1,
	 0xc2, 0x60, 0x3a, 0x23, 0x40, 0x27, 0x3d, 0x22,

	0xc3, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	 0x68, 0x69, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
	0xca, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70,
	 0x71, 0x72, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0,
	0xd1, 0x7e, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
	 0x79, 0x7a, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
	 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
	0x7b, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	 0x48, 0x49, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed,
	0x7d, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50,
	 0x51, 0x52, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3,
	0x5c, 0x9f, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
	 0x59, 0x5a, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	 0x38, 0x39, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff};

char	*getenv();
static	void	title(), prt_lbl(), search(), enquire();
static	void	hdr_chk(), eof_chk(), mt_fskip(), mt_mskip(), mt_bskip();
static	int	chk_lbl();

void	main(argc,argv,envp)
int	argc;
char	*argv[], *envp[];
{
	int	cond;
	char	*devname;

	if( argc > 2 ){
		fputs(argv[0],stdout);
		fputs(" : too many arguments\n",stdout);
		exit(8);
	}
	else if( argc == 2 ){
		devname = argv[1];
	}
	else{
		devname = getenv("TAPE");
		if( devname == NULL )devname = DEFTAPE;
	}

	fhd = open(devname,O_RDONLY);
	if( fhd == -1 ){
		fputs(devname,stdout);
		fputs(" : Device Open Error\n",stdout);
		exit(16);
	}

	/* Check the Start Position */
	if(( cond = chk_lbl(label1)) == 4 ){
		if( chk_lbl(label2) != 5 ){
			fputs("File End Label incomplete\n",stdout);
			mt_bskip(-1);	mt_bskip(-1);
			exit(4);
		}
		fputs("File at the End Label of File '",stdout);
		label1[21] = 0;	fputs(label1+4,stdout);
		fputs("' : Skipping \n",stdout);
		mt_fskip(1);
		cond = chk_lbl(label1);
	}
	switch( cond ){
	case 1:
		fputs("Tape at Volume Label : Volume Serial is '",stdout);
		label1[10] = 0;	fputs(label1+4,stdout);
		fputs("'\n",stdout);
		fputs("File Number to be Located ? ",stdout);
		gets(label1);	search( atoi(label1) - 1 );
		break;
	case 2:
		hdr_chk();	title();	prt_lbl();
		mt_bskip(-2);
		break;
	case 0:
		mt_fskip(EOMSTOP-2);
		if(( chk_lbl(label1) != 4 ) || ( chk_lbl(label2) != 5 )){
			fputs("Tape at File Mark\n",stdout);
			mt_fskip(2-EOMSTOP);	mt_mskip(-1);
			exit(4);
		}
		fputs("Tape at the Tape End : Last File Name is '",stdout);
		label1[21] = 0;	fputs(label1+4,stdout);
		fputs("'\n",stdout);
		mt_fskip(1);
		break;
	default:
		fputs("Tape not at any Label\n",stdout);
		mt_bskip(-1);
		exit(4);
	}

	/* Main Loop */
	for(;;){
		enquire();
		fputs("File Skip Value ? ",stdout);
		gets(label1);	search(atoi(label1));
	}
}

/* MT Movement */
static	void	search(skip)
int	skip;
{
	title();
	for(;;){
		switch( chk_lbl(label1) ){
		case 1:
			fputs("Return to Tape Head\n",stdout);
			if( chk_lbl(label1) != 2 ){
				fputs("Volume Label not followed by File Header\n",stdout);
				exit(4);
			}
			hdr_chk();	prt_lbl();
			mt_bskip(-2);
			return;
		case 2:
			hdr_chk();	prt_lbl();
			break;
		case 0:
			fputs("Tape at the Tape End\n",stdout);
			mt_fskip(EOMSTOP-2);	mt_fskip(1);
			if( skip >= 0 )return;
			break;
		default:
			fputs("File Header not found\n",stdout);
			mt_bskip(-1);
			exit(4);
		}
		if( skip > 0 ){
			mt_fskip(2);	eof_chk();
			mt_fskip(1);
			skip--;
		}
		else if( skip < 0 ){
			mt_fskip(-1);	eof_chk();
			mt_fskip(-2);
			skip++;
		}
		else /* skip == 0 */{
			mt_bskip(-2);
			return;
		}
	}
}

/* Check the 2nd File Header Label */
static	void	hdr_chk()
{
	if( chk_lbl(label2) != 3 ){
		fputs("File Header incomplete\n",stdout);
		mt_bskip(-1);	mt_bskip(-1);
		exit(4);
	}
}

/* Check if EOF Label Exists */
static	void	eof_chk()
{
	if( chk_lbl(label1) != 4 ){
		fputs("File End Label not found\n",stdout);
		mt_bskip(-1);
		exit(4);
	}
	if( chk_lbl(label2) != 5 ){
		fputs("File End Label incomplete\n",stdout);
		mt_bskip(-1);	mt_bskip(-1);
		exit(4);
	}
}

/* Terminating Procedure */
static	void	enquire()
{
	for(;;){
		fputs("Continue(C),Retain(T) or Locate to File Body(L) ? ",stdout);
		gets(label1);
		switch(label1[0]){
			case 'L':
			case 'l':
				mt_fskip(1);
				/* No Break Here */
			case 'T':
			case 't':
				exit(0);
			case 'C':
			case 'c':
				return;
		}
	}
}

/* Print Title Line for Label Information */
static	void	title()
{
	fputs("DSNO   DATASET NAME   RECFM BLKSIZE LRECL CREATE EXPIRE\n",stdout);
}

/* Print out File Label Information */
static	void	prt_lbl()
{
	int	i;
	char	list[64];

	for( i=0 ; i<55 ; i++ )list[i] = ' ';
	list[55] = '\n';	list[56] = 0;

	memcpy(list   ,label1+31,4);	/* File Serial No. */
	memcpy(list+5 ,label1+4 ,17);	/* File Name */
	memcpy(list+23,label2+4 ,1);	/* Record Form */
	memcpy(list+24,label2+38,1);
	memcpy(list+25,label2+36,1);
	memcpy(list+29,label2+5 ,5);	/* Block Size */
	memcpy(list+36,label2+10,5);	/* Logical record Length */
	memcpy(list+42,label1+41,6);	/* Creation Date */
	memcpy(list+49,label1+47,6);	/* Expiration Date */
	fputs(list,stdout);
	return;
}

/* Check Label Length and ID */
static	int	chk_lbl(buffer)
unsigned char	*buffer;
{
	int	i, len;
	unsigned char	id[5];

	len = read(fhd,buffer,81);
	if( len == 0 )return 0;
	for( i=0 ; i<80 ; i++ )
		buffer[i] = ea_table[buffer[i]];
	for( i=len ; i<81 ; i++ )
		buffer[i]= 0;
	if( len != 80 )return -1;
	for( i=0 ; i<4 ; i++ )
		id[i] = buffer[i];
	id[4] = 0;
	if( strcmp(id,"VOL1") == 0 )return 1;
	if( strcmp(id,"HDR1") == 0 )return 2;
	if( strcmp(id,"HDR2") == 0 )return 3;
	if( strcmp(id,"EOF1") == 0 )return 4;
	if( strcmp(id,"EOF2") == 0 )return 5;
	return -1;
}

static	void	mt_fskip(i)
int	i;
{
#if	TMSKIP
	char	buffer[65535];
#endif

	mtopbuf.mt_count = ( i > 0 ) ? i : TMSKIP-i;
	mtopbuf.mt_op = ( i > 0 ) ? MTFSF : MTBSF;
	ioctl(fhd,MTIOCTOP,&mtopbuf);

#if	TMSKIP
	if( i < 0 ){
		if( read(fhd,buffer,65535) != 0 )mt_bskip(-1);
	}
#endif
}

static	void	mt_mskip(i)
int	i;
{
	mtopbuf.mt_count = ( i > 0 ) ? i : -i;
	mtopbuf.mt_op = ( i > 0 ) ? MTFSF : MTBSFM;
	ioctl(fhd,MTIOCTOP,&mtopbuf);
}

static	void	mt_bskip(i)
int	i;
{
	mtopbuf.mt_count = ( i > 0 ) ? i : -i;
	mtopbuf.mt_op = ( i > 0 ) ? MTFSR : MTBSR;
	ioctl(fhd,MTIOCTOP,&mtopbuf);
}
